# A script to build annotation data packages.
# pkgName - a character string for the name of the data package to be
#           built (e. g. hgu95a, rgu34a)
# organism - a character string for the name of the organism of
#            concern (now can only be "human", "mouse", or "rat")
# pkgPath - a character string for the full path of an existing
#           directory where the built backage will be stored
# version - a character string for the version number
# baseName - a character string for the name of a file to be used as a
#            base file to base source data.
# baseMapType - a character string that is either "gb" or "ug" to
#             indicate whether the probe ids in baseName are mapped to
#             GenBack accession numbers or UniGene ids.
# otherSrc - a vector of named character strings for the names of files that
#          contain mappings between probe ids of baseName and
#          LobusLink ids that will be used to obtain the unified
#          mappings between probe ids of baseName and LocusLink ids
#          based on all the sources. The strings should not contain
#          any number.
# makeXML - a boolean to indicate whether an XML version will also be
#         generated.
# srcUrls - a vector of names character strings for the urls where
#         source data files will be retained. Valid sources are LocusLink,
#         UniGene, Golden Path, Gene Ontology, and KEGG. The names for the
#         character strings should be LL, UG, GP, GO, and KEGG,
#         respectively. LL and UG are required.
# author - a named vector of character string with a name element for
#          the name of the author and address element for the email
#          address of the author.
#
# This function remains to be long to avoid passing large data files
# around considering the size of the data files being processed.
#
# Copyright 2003, J. Zhang, all rights reserved.
#

ABPkgBuilder <- function(baseName, srcUrls, baseMapType = c("gb",
                         "ug", "ll", "image"),
                         otherSrc = NULL, pkgName, pkgPath,
                         organism = c("human", "mouse", "rat"),
                         version = "1.1.0", makeXML = TRUE,
                         author = list(author = "who",
                         maintainer = "who@email.com"), fromWeb = TRUE){
    baseMapType <- match.arg(baseMapType)
    switch(baseMapType,
           gb = baseParser <- getBaseParsers("gb"),
           ug = baseParser <- getBaseParsers("ug"),
           image = baseParser <- getBaseParsers("image"),
           ll = baseParser <- getBaseParsers("ll"),
           stop("Invalid imput for baseMapType"))
    organism <- match.arg(organism)
    switch(organism,
           human =,
           mouse =,
           rat = TRUE,
           stop("Invalid input for organism"))
    print("It may take me a while to process the data. Be patient!")
    # Force upper letters
    names(srcUrls) <- toupper(names(srcUrls))
    if(is.null(srcUrls["LL"]) || is.null(srcUrls["UG"]) ||
       is.null(srcUrls["GO"])){
        stop("Source ulrs for LocusLink, UniGene, and GO are required!")
    }
    # Write AnnInfo to .GlobalEnv to make Annotation information available
    makeSrcInfo()
    # Set the path to parsers and base files
    path <- file.path(.path.package("AnnBuilder"), "data")
    # Get the parsers to map probe ids to LL ids through GB or UG ids
 #   baseParsers <- getBaseParsers(baseMapType)
    # Read in the base file
    options(show.error.messages = FALSE)
    base <- try(matrix(scan(baseName, what = "", sep = "\t", quote = "",
                        quiet = TRUE), ncol = 2, byrow = TRUE))
    options(show.error.messages = TRUE)
    if(inherits(base, "try-error")){
        stop(paste("Base file", baseName,
                   "may not be valid or have two columns"))
    }
    colnames(base) <- c("PROBE", "ACC")
    # Need to get source data from LocusLink(LL), UniGene(UG),
    # GoldenPath(GP), Gene Ontology(GO), and KEGG(KEGG). Instantiate
    # objects for each of them. Default values were right at the time
    # of writing but may be wrong later. Make corrections as needed.
    ll <- LL(srcUrl = srcUrls["LL"],
             parser = baseParser["LL"], baseFile = baseName)
    ug <- UG(srcUrl = srcUrls["UG"],
             parser = baseParser["UG"], baseFile = baseName,
             organism = organism)
    go <- GO(srcUrl = srcUrls["GO"])
    if(!is.na(srcUrls["GP"])){
        gp <- GP(srcUrl = srcUrls["GP"], organism = organism)
    }
    if(!is.na(srcUrls["KEGG"])){
        kegg <- KEGG(srcUrl = srcUrls["KEGG"], organism = organism)
    }
    if(baseMapType != "ll"){
        if(baseMapType == "image"){
            unified <- unifyMappings(base, NULL, ug, otherSrc, fromWeb)
        }else{
            unified <- unifyMappings(base, ll, ug, otherSrc, fromWeb)
        }
    }else{
        unified <- baseName
    }
    # Using the unified mapping as the base file, the data file
    # (ll_tmpl) can be parsed to get annoation data from LocusLink
    # using the correct parser
    parser(ll) <- file.path(path, "llParser")
    baseFile(ll) <- unified
    options(show.error.messages = FALSE)
    annotation <- try(parseData(ll, ncol = 15, fromWeb = fromWeb,
                         mergeKey = FALSE))
    options(show.error.messages = TRUE)
    if(inherits(annotation, "try-error")){
        stop(paste("Parsing LocusLink data failed because of:\n\n", annotation))
    }
     colnames(annotation) <- c("PROBE", "ACCNUM", "LOCUSID", "UNIGENE",
                                  "GENENAME", "SYMBOL","CHR", "MAP",
                                  "PMID", "GRIF", "SUMFUNC", "GO",
                                  "OMIM", "NM", "NP")
    if(!is.na(srcUrls["GP"])){
        # Get strand data by processing data from GoldenPath
        options(show.error.messages = FALSE)
        strand <- try(getStrand(gp))
        options(show.error.messages = TRUE)
        if(inherits(strand, "try-error")){
            stop(paste("Failed to parse Golden Path data because of:\n\n",
                       strand))
        }
        # An extra merge is need to keep column order of annoation
        # unchanged after merging
#        strand <- merge(annotation[,c("PROBE", "LOCUSID")],
#                    strand, by = "LOCUSID", all.x = TRUE)[,c("PROBE",
#                                            "CHRORI", "CHRLOC")]
        annotation <- merge(annotation, strand, by = "LOCUSID", all.x = TRUE)
    }
    if(!is.na(srcUrls["KEGG"])){
        # Adding KEGG pathway data to annoataion
        pathNEnzyme <- mapLL2ECNPName(kegg)
        annotation <- merge(annotation, pathNEnzyme$llpathname,
                            by = "LOCUSID", all.x = TRUE)
        annotation <- merge(annotation, pathNEnzyme$llec,
                            by = "LOCUSID", all.x = TRUE)
    }
    if(!is.na(srcUrls["HG"])){
        intID2LL <- getLL2IntID(procHomoData(srcUrls["HG"]), organism)
        annotation <- merge(annotation, intID2LL,
                            by = "LOCUSID", all.x = TRUE)

    }
    annotation <- as.matrix(annotation)
    # Convert "" to NA
    annotation[annotation == ""] <- NA
    # Probe id can not be NULL
    annotation <- annotation[!is.null(annotation[, "PROBE"]),]
    if(makeXML){
        multC <- colnames(annotation)[is.element(colnames(annotation),
                                             getMultiColNames())]
        typeC <- colnames(annotation)[is.element(colnames(annotation),
                                             getTypeColNames())]
        XMLOut <- file.path(pkgPath, paste(pkgName, ".xml", sep = ""))
        fileToXML(targetName = pkgName, outName = XMLOut,
                  inName = annotation, colNames = "",
                  idColName = "PROBE", multColNames = multC,
                  typeColNames = typeC, isFile = FALSE, version = version)
    }
    # Create a data package with no data
    createEmptyDPkg(pkgName, pkgPath, force = TRUE)
    # Write data to the package for one to one mappings
    for(i in getUniColNames()){
        env <- new.env(hash = TRUE, parent = NULL)
        if(i == "LOCUSID"){
            multiassign(annotation[,"PROBE"],
                        as.integer(annotation[,i]), env)
        }else{
            multiassign(annotation[,"PROBE"], as.vector(annotation[,i]), env)
        }
        assign(paste(pkgName, i, sep = ""), env)
        save(list = paste(pkgName, i, sep = ""), file =  file.path(pkgPath,
             pkgName, "data", paste(pkgName, i, ".rda", sep = "")))
    }

    # Write data to the package for one to many mappings with
    # multiple mappings separated by a ";"
    for(i in intersect(colnames(annotation), getMultiColNames())){
        env <- new.env(hash = TRUE, parent = NULL)
        multiassign(annotation[,"PROBE"],
               lapply(annotation[,i], splitEntry), env)
        assign(paste(pkgName, i, sep = ""), env)
        save(list = paste(pkgName, i, sep = ""), file =  file.path(pkgPath,
             pkgName, "data", paste(pkgName, i, ".rda", sep = "")))
    }
    # Write the mappings between probe ids and evidence provided by LL
    # and  CHRORI and CHRLOC. All of them have a "@" separating some
    # descriptive values for an entry. Multiple entries are separated
    # by ";"
    for(i in intersect(colnames(annotation), c("GO", "CHRLOC"))){
        env <- new.env(hash = TRUE, parent = NULL)
        if(i == "CHRLOC"){
            multiassign(annotation[, "PROBE"],
                        lapply(annotation[, i], twoStepSplit,
                               asNumeric = TRUE), env)
        }else{
            multiassign(annotation[, "PROBE"],
                        lapply(annotation[, i], twoStepSplit), env)
        }
        assign(paste(pkgName, i, sep = ""), env)
        save(list = paste(pkgName, i, sep = ""),
             file =  file.path(pkgPath, pkgName, "data",
             paste(pkgName, i, ".rda", sep = "")))
    }
    # Get the reverse mapping between PubMed, pathway, enzyme and probe
    # ids. Column names have to be modified for each reverse mapping
    env <- cols2Env(annotation[,c("PMID", "PROBE")],
                    colNames = c("PMID", "PROBE"), keyColName = "PMID")
    assign(paste(pkgName, "PMID2PROBE", sep = ""), env)
    save(list = paste(pkgName, "PMID2PROBE", sep = ""),
         file =  file.path(pkgPath, pkgName, "data",
         paste(pkgName, "PMID2PROBE.rda", sep = "")))
    if(!is.na(srcUrls["KEGG"])){
        env <- cols2Env(annotation[,c("PATH", "PROBE")],
                        colNames = c("PATH", "PROBE"), keyColName = "PATH")
        assign(paste(pkgName, "PATH2PROBE", sep = ""), env)
        save(list = paste(pkgName, "PATH2PROBE", sep = ""),
             file =  file.path(pkgPath, pkgName, "data",
             paste(pkgName, "PATH2PROBE.rda", sep = "")))
        env <- cols2Env(annotation[,c("ENZYME", "PROBE")],
                        colNames = c("ENZYME", "PROBE"),
                        keyColName = "ENZYME")
        assign(paste(pkgName, "ENZYME2PROBE", sep = ""), env)
        save(list = paste(pkgName, "ENZYME2PROBE", sep = ""),
             file =  file.path(pkgPath, pkgName, "data",
             paste(pkgName, "ENZYME2PROBE.rda", sep = "")))
    }
    # Reassign a parser and base file to parse LocusLink data to map
    # GO ids to probe ids
    parser(ll) <- file.path(path, "GO2ProbeParser")
    options(show.error.messages = FALSE)
    go2Probe <- try(parseData(ll, ncol = 3, fromWeb = fromWeb))
    options(show.error.messages = TRUE)
    if(inherits(go2Probe, "try-error")){
        stop(paste("Failed to parse LocusLink data because of:\n\n",
                   go2Probe))
    }
    if(is.null(go2Probe)){
        warning("No mappings between GO and probe ids found")
    }else if(nrow(go2Probe) == 1){
        go2Probe <- matrix(go2Probe[, 1:2], nrow = 1)
        colnames(go2Probe) <- c("GO", "GO2PROBE")
        env <- new.env(hash = TRUE, parent = NULL)
        assign(go2Probe["GO"], go2Probe["GO2Probe"], env)
    }else{
        colnames(go2Probe) <- c("GO", "GO2PROBE", "COUNTS")
        # Drop column counts for now. Will get ride of the extra column
        go2Probe <- go2Probe[, c("GO", "GO2PROBE")]
        # Remove "" for GO id resulting from mapping probe ids to NA GO id
        go2Probe <- go2Probe[go2Probe[,1] != "", ]
        # Write GO id to probe id mappings to data package
        env <- new.env(hash = TRUE, parent = NULL)
        multiassign(as.vector(go2Probe[, 1]),
                    sapply(go2Probe[, 2], splitEntry), env)
        assign(paste(pkgName, "GO2PROBE", sep = ""), env)
        save(list = paste(pkgName, "GO2PROBE", sep = ""),
             file =  file.path(pkgPath, pkgName, "data",
             paste(pkgName, "GO2PROBE.rda", sep = "")))
    }
    # Get the mappings between GO ids to all the probe ids (including
    # children)
    options(show.error.messages = FALSE)
    goData <- try(readData(go, xml = TRUE, fromWeb = fromWeb))
    options(show.error.messages = TRUE)
    go2All <- mapGO2AllProbe(go2Probe, goData, "", sep = ";", all = TRUE)
    # Write GO id to all probe ids mappings to data package
    env <- new.env(hash = TRUE, parent = NULL)
    multiassign(names(go2All), sapply(go2All, splitEntry), env)
    assign(paste(pkgName, "GO2ALLPROBES", sep = ""), env)
    save(list = paste(pkgName, "GO2ALLPROBES", sep = ""),
         file =  file.path(pkgPath, pkgName, "data",
         paste(pkgName, "GO2ALLPROBES.rda", sep = "")))
    go2All <- cbind(names(go2All), go2All)
    colnames(go2All) <- c("GO", "GO2ALLPROBES")
    if(makeXML){
        if(!is.null(go2All) && !is.null(go2Probe)){
            mergedGO <- as.matrix(merge(as.matrix(go2All),
                          as.matrix(go2Probe), by = "GO", all.x = TRUE))
            XMLByNum <- file.path(pkgPath, paste(pkgName, "ByNum.xml",
                                                 sep = ""))
#            fileToXML("GOByNum", XMLByNum, mergedGO, "GO",
#                      c("GO", "GO2ALLPROBES", "GO2PROBE"),
#                      c("GO2PROBE","GO2ALLPROBES"), "", ";", "",
#                      organism = "GO", isFile = FALSE, version = version)
            fileToXML(targetName = "GOByNum", outName = XMLByNum,
                  inName = mergedGO, colNames = "",
                  idColName = "GO", multColNames = c("GO2ALLPROBES",
                                    "GO2PROBE"),
                  typeColNames = "", isFile = FALSE, version = version)
        }
    }
    writeAccessory(pkgName, pkgPath, organism, version, author)
    # Write the quality control data
    # Quality control
    getDPStats(baseName, pkgName, pkgPath)
    writeMan4QC(pkgName, pkgPath)
    # Clean up
    unlink(unified)
}


unifyMappings <- function(base, ll, ug, otherSrc, fromWeb){
    trusted <- NULL
    # Get the unified mappings between probe ids and locusLink ids
    # based on multiple sources.
    if(!is.null(ll)){
        options(show.error.messages = FALSE)
        llMapping <- try(parseData(ll, fromWeb = fromWeb))
        options(show.error.messages = TRUE)
        if(inherits(llMapping, "try-error")){
            stop(paste("Failed to get or parse LocusLink data",
                       "because of:\n\n",llMapping))
        }
        colnames(llMapping) <- c("PROBE", "LL")
    }
    if(!is.null(ug)){
        options(show.error.messages = FALSE)
        ugMapping <- try(parseData(ug, fromWeb = fromWeb))
        options(show.error.messages = TRUE)
        if(inherits(ugMapping, "try-error")){
            stop(paste("Failed to get or parse UniGene data becaus of:\n\n",
                       ugMapping))
        }
        colnames(ugMapping) <- c("PROBE", "UG")
    }
    # Merge data from all the sources by probe id based on the base
    # file. The probe id is the first column ("V1") in this case.
    # Merge two at a time.
    if(!is.null(ll) && nrow(llMapping) > 0){
        merged <- merge(base, llMapping, by = "PROBE", all.x = TRUE)
        trusted <- "LL"
         # Free the space
        llMapping <- NULL
    }else{
        merged <- base
    }
    if(!is.null(ug) && nrow(ugMapping) > 0){
        merged <- merge(merged, ugMapping, by = "PROBE", all.x = TRUE)
        trusted <- c(trusted, "UG")
        ugMapping <- NULL
    }
    # The following three mappings are for Affymetrix gene chips.
    # Skip if the target of annotation is not a Affymetrix chip with
    # existing mapping data file in the data directory of AnnBuilder.
    if(!is.null(otherSrc)){
        for(i in names(otherSrc)){
            options(show.error.messages = FALSE)
            temp <- try(matrix(scan(otherSrc[i], what = "", sep ="\t",
                         quote = "", quiet = TRUE), ncol = 2, byrow = TRUE))
            options(show.error.messages = TRUE)
            if(inherits(temp, "try-error")){
                stop(paste("File", otherSrc[i],
                           "may not be valid or have two columns"))
            }
            colnames(temp) <- c("PROBE", i)
            merged <- merge(merged, temp, by = "PROBE", all.x = TRUE)
        }
    }
    # Change "NA" tp NA
    merged <- as.matrix(merged)
    merged[merged == "NA"] <- NA
    # Get the unified mapping based on all the sources used
    unified <- resolveMaps(merged, trusted = trusted,
                            srcs = c(names(otherSrc)))
    return(unified)
}

getBaseParsers <- function(baseMapType = c("gb", "ug", "image", "ll")){
    type <- match.arg(baseMapType)
    path <- file.path(.path.package("AnnBuilder"), "data")
    switch(type,
           gb = return(c(LL = file.path(path, "gbLLParser"),
                         UG = file.path(path, "gbUGParser"))),
           ug = return(c(LL = file.path(path, "ugLLParser"),
                         UG = file.path(path, "ugUGParser"))),
           image = return(c(UG = file.path(path, "imageUGParser"),
                          LL = "NA")),
           ll = return(c(LL = file.path(path, "llParser"),
                         UG = "NA")),
           stop("Invalid type"))
}

createEmptyDPkg <- function(pkgName, pkgPath,
                            folders = c("man", "R", "data"), force = TRUE){
    if(file.exists(file.path(pkgPath, pkgName))){
        if(!force){
            stop(paste("Package", pkgName, "already exists"))
        }else{
            unlink(file.path(pkgPath, pkgName), TRUE)
        }
    }
    dir.create(file.path(pkgPath, pkgName))
    for(i in folders){
        dir.create(file.path(pkgPath, pkgName, i))
    }
}
# Split multiple entry for a given mapping
splitEntry <- function(dataRow, sep = ";", asNumeric = FALSE){
    if(is.na(dataRow) || is.null(dataRow) || dataRow == ""){
        return(NA)
    }else{
        if(asNumeric){
            return(as.numeric(unlist(strsplit(dataRow, sep))))
        }else{
            return(unlist(strsplit(dataRow, sep)))
        }
    }
}
# Split multiple entry with two separaters (e. g. 12345@18;67891@18)
twoStepSplit <- function(dataRow, entrySep = ";", eleSep = "@",
    asNumeric = FALSE){
    splitEle <- function(entry){
        if(!is.na(entry)){
            temp <- unlist(strsplit(entry, eleSep), use.names = FALSE)
            if(asNumeric){
                elements <- as.numeric(temp[1])
            }else{
                elements <- temp[1]
            }
            names(elements) <- temp[2]
            return(elements)
        }else{
            return(NA)
        }
    }
    temp <- unlist(strsplit(dataRow, entrySep), use.names = FALSE)
    # Force "NA" to be NA
    temp[temp == "NA"] <- NA
    return(unlist(lapply(temp, splitEle), use.names = TRUE))
}

# This function takes two data columns from a data file (e. g. two
# columns) and generates an environment object with
# values in one of the columns as keys and values in the other as
# values. A given row in either column may have multiple values
# separated by a separator(e. g. "a;b") and has to be dealt with.
#
# Copyright 2002, J. Zhang, all rights reserved
#

cols2Env <- function(cols, colNames, keyColName = colNames[1], sep = ";"){

    # Environment to return
    dataEnv <- new.env(hash = TRUE, parent = NULL)

    # Force a matrix
    cols <- as.matrix(cols)
    if(!missing(colNames)){
        colnames(cols) <- colNames
    }

    if(ncol(cols) != 2){
        stop("The input data do not have two columns")
    }
    # Determine the column name
    valColName <- colNames[colNames != keyColName]
    # Match cols by finding matches for all the potential multiple
    # matches
    cols <- matchAll(cols, keyColName)
    colnames(cols) <- colNames
    #Get unduplicated row
    unDups <- cols[!duplicated(cols[,keyColName]),]
    colnames(unDups) <- colNames
    # Write unique keys and values
    multiassign(unDups[, keyColName], unDups[, valColName], dataEnv)
    # Process the duplications
    dups <- cols[duplicated(cols[,keyColName]),]
    colnames(dups) <- colNames
    if(nrow(dups) > 0){
        for(i in 1:nrow(dups)){
            assign(dups[i,keyColName], unique(c(get(dups[i, keyColName],
                                   dataEnv), dups[i,valColName])), dataEnv)
        }
    }

    return(dataEnv)
}

matchAll <- function(cols, keyColName){
    matched <- NULL

    for(i in 1:nrow(cols)){
        temp <- matchOneRow(cols[i,], keyColName)
        if(!is.null(temp)){
            if(is.null(matched)){
                matched <- temp
            }else{
                matched <- rbind(matched, temp)
            }
        }
    }
    return(matched)
}

matchOneRow <- function(cols, keyColName, sep = ";"){

    doMatch <- function(col1id){
        if(is.null(matched)){
            matched <<- cbind(col1id, col2)
        }else{
            matched <<- rbind(matched, cbind(col1id, col2))
        }
    }

    matched <- NULL
    # Key can not be any of these
    if(cols[keyColName] == "" || is.na(cols[keyColName])||
       is.null(cols[keyColName])){
        return(NULL)
    }

    col1 <- unlist(strsplit(cols[1], sep), use.names = FALSE)
    col2 <- unlist(strsplit(cols[2], sep), use.names = FALSE)
    # If no mapping is found for key, assign NA to key
    if(match(keyColName, names(cols)) == 1 && length(col2) == 0){
        return(cbind(col1, NA))
    }else if(match(keyColName, names(cols)) == 2 && length(col1) == 0){
        return(cbind(NA, col2))
    }
    # Otherwise, do a full matching
    sapply(col1, doMatch)

    return(matched)
}

# This function creates descriptive statistics for a given data
# package.
#
# Copyright 2002, J. Zhang. All rights reserved.
#

getDPStats <- function(baseF, pkgName, pkgPath, saveList = TRUE,
                       isFile = TRUE){

    dateBuilt <- getDate(pkgName, pkgPath)
    keyMapped <- getProbeNum(pkgName, pkgPath, TRUE)
    probeBased <- paste(pkgName, getPBased(), sep = "")

    otherMapped <- NULL
    for(i in setdiff(names(keyMapped), probeBased)){
        if(any(i == names(keyMapped))){
            tempNum <- as.numeric(keyMapped[i])
            names(tempNum) <- i
            otherMapped <- c(otherMapped, tempNum)
        }
    }

    if(!is.null(baseF) && !is.na(baseF) && baseF != ""){
        if(isFile){
            srcNumProbes <- length(readLines(baseF))
        }else{
            srcNumProbes <- nrow(baseF)
        }
        numProbes <- getProbeNum(pkgName, pkgPath, FALSE)
        keyMissMatch <- matchProbes(baseF, pkgName, pkgPath,
                                    probeBased, isFile = isFile)

        numMissMatch <- "None"
        for(i in names(numProbes)){
            if(any(i == probeBased) && numProbes[i] != srcNumProbes){
                if(numMissMatch == "None"){
                    numMissMatch <- NULL
                }
                numMissMatch <- c(numMissMatch, i)
            }
        }

        probeMapped <- NULL
        for(i in names(keyMapped)){
            if(any(i == probeBased)){
                tempNum <- as.numeric(keyMapped[i])
                names(tempNum) <- i
                probeMapped <- c(probeMapped, tempNum)
            }
        }
    }else{
        probeMapped <- NULL
        numMissMatch <- NULL
        srcNumProbes <- NULL
        numProbes <- NULL
        keyMissMatch <- NULL
    }
    statsList <- list(name = pkgName,
                      built = dateBuilt,
                      probeNum = srcNumProbes,
                      numMissMatch = numMissMatch,
                      probeMissMatch = keyMissMatch,
                      probeMapped = probeMapped,
                      otherMapped = otherMapped)
    QCList <- formatABQCList(statsList)
    if(saveList){
        assign(paste(pkgName, "QC", sep = ""), QCList)
        save(list = paste(pkgName, "QC", sep = ""),
             file = file.path(pkgPath, pkgName, "data",
             paste(pkgName, "QC.rda", sep = "")))
    }
    return(statsList)
}

formatABQCList <- function(x){
    QCList <- paste("\n\nQuality control information for ", x$name,
                    "\nDate built:", x$built,
                    ifelse(is.null(x$probeNum), "",
                           paste("\nNumber of probes:", x$probeNum)),
                    ifelse(is.null(x$numMissMatch), "",
                           paste("\nProbe number missmatch:",
                                 paste(x$numMissMatch, sep = "",
                                       collapse = "; "))),
                    ifelse(is.null(x$probeMissMatch), "",
                           paste("\nProbe missmatch:",
                                 paste(x$probeMissMatch, sep = "",
                                       collapse = "; "))),
                    ifelse(is.null(x$probeMapped), "",
                           paste("\nMappings found for ",
                                 "probe based rda files: \n",
                                 paste("\t", names(x$probeMapped),
                                       "found", x$probeMapped,
                                       "of", x$probeNum, sep = " ",
                                       collapse = "\n"),
                                 sep = "", collapse ="\n")),
                    "\nMappings found for non-probe based rda files:\n",
                    paste("\t", names(x$otherMapped), "found",
                          x$otherMapped, sep = " ", collapse = "\n"),"\n\n")
    return(QCList)
}

getDate <- function(pkgName, pkgPath, fromDesc = TRUE){
    if(fromDesc){
        fileName <- file.path(pkgPath, pkgName, "DESCRIPTION")
        if(!file.exists(fileName))
            stop(paste("DESCRIPTION file was not found in",
                       file.path(pkgPath, pkgName), "!"))

        tempFile <- readLines(fileName)

        for (i in tempFile){
            if(regexpr("Created:", i) > 0){
                return(gsub("^Created:(*)", "\\", i))
            }
        }
        return("Date built not found")
    }else{
        return(date())
    }
}

getProbeNum <- function(pkgName, pkgPath, noneNA = FALSE){
    toReturn <- NULL

    dataFolder <- file.path(pkgPath, pkgName, "data")
    rdaFiles <- getDirContent(dataFolder, paste(pkgName, "QC.rda", sep = ""))

    for(i in rdaFiles){
        temp <- unlist(strsplit(i, "\\."))
        load(file.path(dataFolder, i))
        if(!noneNA){
            num <- length(ls(get(temp[1])))
        }else{
            if(length(ls(get(temp[1]))) == 0){
                num <- 0
            }else{
                values <- multiget(ls(get(temp[1])), env = get(temp[1]))
                values <- values[!is.na(values)]
                values <- values[!is.null(values)]
                values <- values[values != ""]
                values <- values[values != "NA"]
                num <- length(values)
            }
        }
        names(num) <- temp[1]
        toReturn <- c(toReturn, num)
    }
    return(toReturn)
}

getDirContent <- function(dirName, exclude = NULL){
    if(is.null(exclude)){
        return(list.files(dirName))
    }else{
        temp <- list.files(dirName)
        index <- unlist(lapply(exclude, grep, temp))
        if(length(index) > 0){
            return(temp[-index])
        }else{
            return(temp)
        }
    }
}

matchProbes <- function(baseF, pkgName, pkgPath, toMatch, isFile = TRUE){

    toReturn <- "None"

    if(isFile){
        srcProbes <- read.delim(baseF, header = FALSE, sep = "\t")
    }else{
        srcProbes <- baseF
    }
    dataFolder <- file.path(pkgPath, pkgName, "data")
    rdaFiles <- list.files(dataFolder)

    for(i in rdaFiles){
        if(regexpr(".rda", i) > 0){
            name <- gsub(paste(pkgName,"(.*).rda", sep = ""),
                         "\\1", i)
            load(file.path(dataFolder, i))
            if(any(toMatch == name)){
                if(any(is.element(srcProbes[,1],
                                  ls(get(temp[1]))) == FALSE)){
                    if(toReturn == "None"){
                        toReturn <- NULL
                    }
                    toReturn <- c(toReturn, i)
                }
            }
        }
    }
    return(toReturn)
}

getPBased <- function(){

    probeBased <- NULL
    for(i in ls(AnnInfo)){
        if(get(i, AnnInfo)$pbased == "Y"){
            probeBased <- c(probeBased, i)
        }
    }
    return(probeBased)
}

# Returns a vector for the names of columns that have multiple entries
# separated by a ";" in some of the cells
getMultiColNames <- function(){
     return(c("PMID", "PATH", "ENZYME", "CHR", "UNIGENE", "HGID", "OMIM"))
}

getUniColNames <- function(){
    return(c("ACCNUM", "LOCUSID", "GENENAME", "SYMBOL", "MAP",
             "GRIF", "SUMFUNC", "NM", "NP"))
}

getTypeColNames <- function(){
    return(c("GO", "CHRLOC"))
}

#nameGOByCategory <- function(goids, goCategory, sep = ";"){
#    nameGOID <- function(goids){
#        if(is.null(goids) || is.na(goids)){
#            return(goids)
#        }
#        goids <- gsub("@[a-z]*", "", goids, TRUE)
#        goids <- unlist(strsplit(goids, sep), use.names = FALSE)
#        names(goids) <- goCategory[is.element(goCategory[,1], goids), 2]
#        return(goids)
#    }
#    goids <- lapply(goids, nameGOID)
#    goids[goids == "NA"] <- NA
#    return(goids)
#}

# This function writes the organism data
writeOrganism <- function(pkgName, pkgPath, organism){
    assign(paste(pkgName, "ORGANISM", sep = ""), organism)
    save(list = paste(pkgName, "ORGANISM", sep = ""),
             file =  file.path(pkgPath, pkgName, "data",
             paste(pkgName, "ORGANISM.rda", sep = "")))
}

# This function writes chromosome length data
writeChrLength <- function(pkgName, pkgPath, chrLengths){
    assign(paste(pkgName, "CHRLENGTHS", sep = ""), chrLengths)
    save(list = paste(pkgName, "CHRLENGTHS", sep = ""),
             file =  file.path(pkgPath, pkgName, "data",
             paste(pkgName, "CHRLENGTHS.rda", sep = "")))
}

# This function estimates the total lengths of chromosomes by finding
# out the maximum chromosome loaction and then increasing the number by 1000
findChrLength <- function(organism, srcUrl = getSrcUrl("GP", organism)){
    chrLengths <- vector()

    estimateLength <- function(chroNum){
        locs <- locatData[locatData[,1] == as.character(chroNum),]
        if(nrow(locs) == 0){
            temp <- NA
        }else if (nrow(locs) == 1){
            temp <- as.numberic(locs[, 2])
        }else {
            temp <- max(as.numeric(locs[, 2]))
        }
        chrLengths[as.character(chroNum)] <<- temp
    }
    locatData <- getGPData(paste(srcUrl, "refGene.txt.gz", sep = ""),
                           ncol = 10, keep = c(2,4))
    # Remove "chr" from chromosome numbers
    locatData[,1] <- gsub("^chr", "\\", locatData[, 1])
    chromosomes <- unique(locatData[,1])
    # Do numbered chromosomes first
    options(warn = -1)
    numbered <- chromosomes[!is.na(as.numeric(chromosomes))]
    options(warn = 0)
    tt <- sapply(sort(as.numeric(numbered)), estimateLength)
    # Do others
    switch(toupper(organism),
           MOUSE = ,
           RAT = ,
           HUMAN = sapply(c("X", "Y"), estimateLength),
           stop(paste("Do not know what to do with", organism)))
    # Remove entries with random locations
    # Add 1000 bases
    return(chrLengths + 1000)
}
