# Functions that return the correct URL that is going to be used to
# get annotation data from the desirable public data source.
#
# Copyright 2002, J. Zhang, all rights reserved.
#

getSrcUrl <- function(src = "LL", organism = "human", xml = TRUE,
                      dateOnly = FALSE){
    load(file.path(.path.package("AnnBuilder"), "data",
                    "sourceURLs.rda"), .GlobalEnv)
    switch(toupper(src),
           "LL" = return(getLLUrl()),
           "GP" = return(getUCSCUrl(organism)),
           "UG" = return(getUGUrl(organism)),
           "GO" = return(getGOUrl(xml, dateOnly)),
           "KEGG" = return(getKEGGUrl()),
           "GEO" = return(getGEOUrl()),
           "YG" = return(getYGUrl()),
           "HG" = return(getHGUrl()),
           "ALL" = return(getAllUrl(organism)),
           stop(paste("Source", src, "is not supported.",
                      "Has to be LL, GP, UG, GO, KEGG, ALL")))
}

getAllUrl <- function(organism){
    urls <- c(LL = getLLUrl(), GP = getUCSCUrl(organism),
              UG = getUGUrl(organism), GO = getGOUrl(), KEGG = getKEGGUrl(),
              YG = getYGUrl(), HG = getHGUrl())
    return(urls)
}

getYGUrl <- function(){
    return("http://www.yeastgenome.org/DownloadContents.shtml")
}

getLLUrl <- function(){
    return(sourceURLs[["LL"]])
}

getUCSCUrl <- function(organism){
    switch(toupper(organism),
           "HUMAN" = key <- "Annotation database",
           "MOUSE" = key <- "goldenPath/mm.*Annotation database",
           "RAT" = key <- "goldenPath/rn.*Annotation database",
           stop (paste("Organism", organism, "is not supported.")))

    con <- url(sourceURLs[["GP"]], open = "r")
    htmlPage <- readLines(con)
    close(con)
    aLine <- htmlPage[grep(key, htmlPage)[1]]
    got <- gsub(".*(goldenPath/.*/database/).*", "\\1", aLine)
    return(paste(gsub("(.*/).*.html", "\\1", sourceURLs[["GP"]]),
                 got, sep = ""))
}

getUGUrl <- function(organism){
    switch(toupper(organism),
           "HUMAN" = return(paste(sourceURLs[["UG"]], "/Hs.data.gz",
                            sep = "")),
           "MOUSE" = return(paste(sourceURLs[["UG"]], "/Mm.data.gz",
                            sep = "")),
           "RAT" = return(paste(sourceURLs[["UG"]], "/Rn.data.gz", sep = "")),
           stop(paste("Organism", organism, "is not supported.")))
}

# Figures out the url for the latest version of the GO data file
getGOUrl <- function(xml = TRUE, dateOnly = FALSE){
    # Read the html page to get the available data for different builds
    con <- url(sourceURLs[["GO"]], open = "r")
    srcHtml <-  readLines(con)
    close(con)
    builds <- srcHtml[grep("^<IMG SRC=.*>[0-9]*-[0-9]*-[0-9]*/.*", srcHtml)]
    # Loop through the builds to get the latest one
    builds <- sort(gsub(".*>([0-9]*-[0-9]*-[0-9]*)/.*", "\\1",  builds),
                   decreasing = TRUE)
    index <- 1
    # Get the name of the available file in the latest build
    repeat{
        con <- url(paste(sourceURLs[["GO"]], "/", builds[index],
                         sep = ""), open = "r")
        files <- readLines(con)
        close(con)
        if(xml){
            fileName <- gsub(".*(go_[0-9]*-termdb.xml.gz).*",
                    "\\1", files[grep(".*termdb.xml.gz", files)])
        }else{
            fileName <- gsub(".*(go_[0-9]*-termdb-data.gz).*", "\\1",
                             files[grep(".*termdb-dat", files)])
        }
        if(length(fileName) == 0){
            if(index < length(builds)){
                index <- index + 1
            }else{
                stop(paste("The target data file termdb.xxx.gz is not",
                           "available at", paste(sourceURLs[["GO"]],
                                                 "/", latest, sep = "")))
            }
        }else{
            if(dateOnly){
                return(builds[index])
            }else{
                return(paste(sourceURLs[["GO"]], "/",
                             builds[index], "/",fileName, sep = ""))
            }
        }
    }
}

getKEGGUrl <- function(){
    return(sourceURLs[["KEGG"]])
}

getGEOUrl <- function(){
    return(sourceURLs[["GEO"]])
}

getHGUrl <- function(){
    return(sourceURLs[["HG"]])
}
