# This function builds a data package that maps internal HomoloGene
# ids of an organism to LocusLink ids, UniGene ids, percent identity
# of the alignment, and type of similarities of organisms of all
# pairwise best matches based on data from
# "ftp://ftp.ncbi.nih.gov/pub/HomoloGene/hmlg.ftp"
# pkgName - a character string for the name of the data package to be
# created.
# pkgPath - a character string for the path to which the data package
# to be created will be stored.
# version - a character string for the version number of the data
# package to be created.
# author - a list of two elements with the firest one being author - a
# character string for the name of the author of the data package and
# second one being maintainer - a character string for the email of
# the author.
#

homoPkgBuilder <- function(pkgName = "homology", pkgPath, version, author,
                 url = getSrcUrl("HG")){

    homoData <- procHomoData(url)
    makeSrcInfo()
    createEmptyDPkg(pkgName, pkgPath)
    # Write man page templates
    writeManPage(pkgName, pkgPath, "#HGID2LL#", "",
                         src = get("#HGID2LL#", AnnInfo)$src)
    writeManPage(pkgName, pkgPath, "#HGID2GB#", "",
                         src = get("#HGID2GB#", AnnInfo)$src)
    writeManPage(pkgName, pkgPath, "#HGID2PS#", "",
                         src = get("#HGID2PS#", AnnInfo)$src)
#    writeManPage(pkgName, pkgPath, "#HGID2TS#", "",
#                         src = get("#HGID2TS#", AnnInfo)$src)
    writeManPage(pkgName, pkgPath, "#HGID2HGID#", "",
                         src = get("#HGID2HGID#", AnnInfo)$src)
    # Do HGID to LL
    writeRdaNMan(mapIntID(homoData[, c(1, 2, 4, 5, 7, 8)]), pkgName,
                 pkgPath, "HGID2LL")
    # Do HGID to GB
    writeRdaNMan(mapIntID(homoData[, c(1, 2, 6, 5, 9, 8)]), pkgName,
                 pkgPath, "HGID2GB")
    # Do HGID to PS
    mapPS(homoData[, c(1, 2, 3, 4, 5, 7, 8, 10)], pkgName, pkgPath)
    # Do HGID to HGID
    writeRdaNMan(mapIntID(homoData[, c(1, 2, 5, 5, 8, 8)]), pkgName,
                 pkgPath, "HGID2HGID")
    # Remove man page templates
    file.remove(file.path(pkgPath, pkgName, "man",
                           paste(pkgName, "#HGID2LL#.Rd", sep = "")),
                file.path(pkgPath, pkgName, "man",
                           paste(pkgName, "#HGID2GB#.Rd", sep = "")),
                file.path(pkgPath, pkgName, "man",
                           paste(pkgName, "#HGID2PS#.Rd", sep = "")),
                file.path(pkgPath, pkgName, "man",
                           paste(pkgName, "#HGID2HGID#.Rd", sep = "")))
    writeDescription(pkgName, pkgPath, version, author)
    # using pkgName() and the man page for the function
    writeFun (pkgPath, pkgName, organism = "")
    writeMan4Fun(pkgName, pkgPath, "hg")
    # Write .First.lib
    writeZZZ(pkgPath, pkgName)
    # Write the quality control data
    # Quality control
    getDPStats("", pkgName, pkgPath)
    writeMan4QC(pkgName, pkgPath)
}

procHomoData <- function(url = getSrcUrl("HG")){
    # scan() does not work for some reason
    homoData <- as.matrix(read.table(url, header = FALSE, sep = "|",
                           quote = "", as.is = TRUE, strip.white = TRUE,
                           comment.char = ""))
    # remove leading LL from LL ids
    homoData <- matrix(gsub("^LL\\.", "", homoData),
                       ncol = ncol(homoData), byrow = FALSE)
    # make "" NA
    homoData[homoData == ""] <- NA
    # remove leading white spaces
    homoData <- matrix(gsub(" *", "", homoData),
                       ncol = ncol(homoData), byrow = FALSE)
    return(homoData)
}

getLL2IntID <- function(homoData, organism = ""){
    switch(tolower(organism),
           human = org <- "9606",
           mouse = org <- "10090",
           rat = org <- "10116",
           org <- NULL)
    ll2ID <- homoData[, c(4, 5, 1)]
    ll2ID <- rbind(ll2ID, homoData[,c(7, 8, 2)])
    if(!is.null(org)){
        ll2ID <- ll2ID[ll2ID[,3] == org, ]
    }
    ll2ID <- ll2ID[, 1:2]
    # Drop entries without LL
    ll2ID <- ll2ID[ll2ID[, 1] != "NA", ]
    # Get duplicated LL
    if(nrow(ll2ID) > length(unique(ll2ID[,1]))){
        ll2ID <- mergeRowByKey(ll2ID)
    }
    colnames(ll2ID) <- c("LOCUSID", "HGID")

    return(ll2ID)
}

getIntIDMapping <- function(homoData){
    dataEnv <- new.env(hash = TRUE, parent = NULL)
    # Drop entries with empty Internal id or NA for Mapping
    homoData <- homoData[homoData[, 2] != "NA" & !is.na(homoData[,2]) &
                         homoData[, 3] != "NA" & !is.na(homoData[,3]), ]
    # Remove duplication
    homoData <- homoData[!duplicated(homoData),]
    # If there is only one mapping found
    if(is.null(nrow(homoData))){
        homoData <- matrix(homoData, nrow = 1)
    }
    homoData <- cbind(homoData[,2], paste(homoData[,3],
                                      mapOrgs(homoData[,1]), sep = "@"))
    homoData <- mergeRowByKey(homoData, 1)
    if(length(homoData) > 0){
        multiassign(homoData[,1], lapply(homoData[,2], twoStepSplit),
                    env = dataEnv)
    }
    return(dataEnv)
}

mapIntID <- function(homoData){
#    if(combName){
#         IDMapping <- homoData[, c(1, 2, 3, 5, 6, 8)]
#         IDMapping <- rbind(IDMapping, homoData[,c(2, 1, 3, 7, 4, 8)])
#         IDMapping <- cbind(IDMapping[,1],
#                          paste(IDMapping[,2], IDMapping[,5], IDMapping[,3],
#                                  sep = ";"), IDMapping[, c(4, 6)])
#         return(IDMapping)
#    }else{
         IDMapping <- homoData[, c(1, 2, 4, 5)]
         IDMapping <- rbind(IDMapping, homoData[,c(2, 1, 6, 3)])
         return(IDMapping)
#    }
}

mapOrgs <- function(vect){
    for(i in names(orgNameNCode)){
        vect <- gsub(paste("^", i, "$", sep = ""),
                     orgNameNCode[i], vect)
    }
    return(vect)
}

writeRdaNMan <- function(homoData, pkgName, pkgPath, what){
    load(file.path(.path.package("AnnBuilder"), "data", "orgNameNCode.rda"),
         env = .GlobalEnv)
    for(i in names(orgNameNCode)){
        tempList <- list()
        tempData <- homoData[homoData[,1] == i, ]
        if(nrow(tempData) > 0){
            assign(paste(pkgName, i, what, sep = ""),
                   getIntIDMapping(tempData[, 2:4]))
            save(list = paste(pkgName, i, what, sep = ""),
                 file =  file.path(pkgPath, pkgName, "data",
                 paste(pkgName, i, what, ".rda", sep = "")))
            tempList[[what]] <-  paste(i, what, sep = "")
            tempList[["ORGANISM"]] <- orgNameNCode[[i]]
            copySubstitute(file.path(pkgPath, pkgName, "man",
                           paste(pkgName, "#", what, "#.Rd", sep = "")),
                           file.path(pkgPath, pkgName,
                           "man", paste(pkgName, i, what, ".Rd",
                                        sep = "")), tempList, "#")
        }
    }
}

mapPS <- function(homoData, pkgName, pkgPath){
    psMapping <- homoData[, c(1, 2, 3, 5, 6, 8)]
    psMapping <- rbind(psMapping, homoData[,c(2, 1, 3, 7, 4, 8)])
    homoPSs <- apply(psMapping, 1, getHomoPS)

    for(i in names(orgNameNCode)){
        mergeByHGID <- function(){
            mergeOne <- function(hgid){
                PSs <- tempPSs[tempHGIDs == hgid & !is.na(tempHGIDs)]
                noDup[[hgid]] <<- PSs
            }
            noDup <- tempPSs[!duplicated(tempHGIDs) & !is.na(tempHGIDs)]
            names(noDup) <- tempHGIDs[!duplicated(tempHGIDs) &
                                      !is.na(tempHGIDs)]
            trash <- sapply(unique(tempHGIDs[duplicated(tempHGIDs) &
                                            !is.na(tempHGIDs)]), mergeOne)
            return(noDup)
        }
        tempHGIDs <- psMapping[psMapping[,1] == i , 4]
        tempPSs <- homoPSs[psMapping[,1] == i]

        homoPSList <- mergeByHGID()

        if(length(homoPSList) > 0){
            tempEnv <- new.env(hash = TRUE, parent = NULL)
            multiassign(names(homoPSList), homoPSList, tempEnv)
            assign(paste(pkgName, i, "HGID2PS", sep = ""), tempEnv)
            save(list = paste(pkgName, i, "HGID2PS", sep = ""),
                 file =  file.path(pkgPath, pkgName, "data",
                 paste(pkgName, i, "HGID2PS", ".rda", sep = "")))
            tempList[["HGID2PS"]] <-  paste(i, "HGID2PS", sep = "")
            tempList[["ORGANISM"]] <- orgNameNCode[[i]]
            copySubstitute(file.path(pkgPath, pkgName, "man",
                           paste(pkgName, "#HGID2PS#.Rd", sep = "")),
                           file.path(pkgPath, pkgName,
                                   "man", paste(pkgName, i, "HGID2PS.Rd",
                                              sep = "")), tempList, "#")
        }
    }
}

getHomoPS <- function(entries){
    if(length(entries) < 6){
        stop("Incorrect argument length. Must be at least 6")
    }
    if(any(entries[3] == c("B", "b"))){
        return(new("homoPS", psOrg = mapOrgs(entries[2]), psLL = entries[5],
                   psType = entries[3], ps = as.numeric(entries[6])))
    }else{
        return(new("homoPS", psOrg = mapOrgs(entries[2]), psLL = entries[5],
                   psType = entries[3], psURL = entries[6]))
    }
}
